/*
 *      torad - 
 *              Convert an iris image into Radiance floating point picture.
 *
 *                      Greg Ward - 1992
 */

#include "image.h"

/* The following was taken from color.h in ray/src/common of Radiance 2.2 */

#define  RED		0
#define  GRN		1
#define  BLU		2
#define  EXP		3
#define  COLXS		128	/* excess used for exponent */

typedef unsigned char  BYTE;	/* 8-bit unsigned integer */

typedef BYTE  COLR[4];		/* red, green, blue, exponent */

#define  copycolr(c1,c2)	(c1[0]=c2[0],c1[1]=c2[1], \
				c1[2]=c2[2],c1[3]=c2[3])

typedef float  COLOR[3];	/* red, green, blue */

#define  colval(col,pri)	((col)[pri])
#define  setcolor(col,r,g,b)	((col)[RED]=(r),(col)[GRN]=(g),(col)[BLU]=(b))
#define  copycolor(c1,c2)	((c1)[0]=(c2)[0],(c1)[1]=(c2)[1],(c1)[2]=(c2)[2])
#define  scalecolor(col,sf)	((col)[0]*=(sf),(col)[1]*=(sf),(col)[2]*=(sf))
#define  addcolor(c1,c2)	((c1)[0]+=(c2)[0],(c1)[1]+=(c2)[1],(c1)[2]+=(c2)[2])
#define  multcolor(c1,c2)	((c1)[0]*=(c2)[0],(c1)[1]*=(c2)[1],(c1)[2]*=(c2)[2])
#define  COLRFMT		"32-bit_rle_rgbe"

#ifndef ldexp
extern double  ldexp();
#endif
#ifndef atof			/* atof's a macro on some systems! */
extern double  atof();
#endif

/* End of color.h defs */

/* The following was taken from resolu.h in ray/src/common of Radiance 2.2 */

/*
 * Definitions for resolution line in image file.
 *
 * True image orientation is defined by an xy coordinate system
 * whose origin is at the lower left corner of the image, with
 * x increasing to the right and y increasing in the upward direction.
 * This true orientation is independent of how the pixels are actually
 * ordered in the file, which is indicated by the resolution line.
 * This line is of the form "{+-}{XY} xyres {+-}{YX} yxres\n".
 * A typical line for a 1024x600 image might be "-Y 600 +X 1024\n",
 * indicating that the scanlines are in English text order (PIXSTANDARD).
 */

			/* flags for scanline ordering */
#define  XDECR			1
#define  YDECR			2
#define  YMAJOR			4

			/* standard scanline ordering */
#define  PIXSTANDARD		(YMAJOR|YDECR)
#define  PIXSTDFMT		"-Y %d +X %d\n"

			/* structure for image dimensions */
typedef struct {
	int	or;		/* orientation (from flags above) */
	int	xr, yr;		/* x and y resolution */
} RESOLU;

			/* macros to get scanline length and number */
#define  scanlen(rs)		((rs)->or & YMAJOR ? (rs)->xr : (rs)->yr)
#define  numscans(rs)		((rs)->or & YMAJOR ? (rs)->yr : (rs)->xr)

			/* resolution string buffer and its size */
#define  RESOLU_BUFLEN		32
extern char  resolu_buf[RESOLU_BUFLEN];

			/* macros for reading/writing resolution struct */
#define  fputsresolu(rs,fp)	fputs(resolu2str(resolu_buf,rs),fp)
#define  fgetsresolu(rs,fp)	str2resolu(rs, \
					fgets(resolu_buf,RESOLU_BUFLEN,fp))

			/* reading/writing of standard ordering */
#define  fprtresolu(sl,ns,fp)	fprintf(fp,PIXSTDFMT,ns,sl)
#define  fscnresolu(sl,ns,fp)	(fscanf(fp,PIXSTDFMT,ns,sl)==2)

extern char  *fgets(), *resolu2str();

/* End of resolu.h defs */

#define MAXWIDTH	8192

FILE *ofile;

unsigned char outbuf[4*MAXWIDTH];

main(argc,argv)
int argc;
char **argv;
{
    static short r[MAXWIDTH];
    static short g[MAXWIDTH];
    static short b[MAXWIDTH];
    IMAGE *iimage;
    int a = 1;
    int xsize, ysize, zsize;
    int y;

    if(argc<3) {
	fprintf(stderr,"usage: torad [-g gamma] irisimage radimage\n");
	exit(1);
    }
    if (argc>4 && !strcmp(argv[a], "-g")) {
	setcolrgam(atof(argv[a+1]));
	a += 2;
    } else
	setcolrgam(1.0);
    if( (iimage=iopen(argv[a],"r")) == NULL ) {
        fprintf(stderr,"torad: can't open input file %s\n",argv[a]);
        exit(1);
    }
    xsize = iimage->xsize;
    ysize = iimage->ysize;
    zsize = iimage->zsize;
    ofile = fopen(argv[a+1],"w");
    printargs(argc,argv,ofile);
    fputformat(COLRFMT,ofile);
    putc('\n',ofile);
    fprtresolu(xsize,ysize,ofile);
    for(y=0; y<ysize; y++) {
        if(zsize<3) {
            getrow(iimage,r,(ysize-1)-y,0);
            outrow(r,r,r,xsize);
        } else {
            getrow(iimage,r,(ysize-1)-y,0);
            getrow(iimage,g,(ysize-1)-y,1);
            getrow(iimage,b,(ysize-1)-y,2);
            outrow(r,g,b,xsize);
        }
    }
    fclose(ofile);
    exit(0);
}

swizzleColr(rptr,gptr,bptr,cptr,n)
register unsigned short *rptr, *gptr, *bptr;
register unsigned char *cptr;
register int n;
{
    while(n--) {
        *cptr++ = *rptr++;
        *cptr++ = *gptr++;
        *cptr++ = *bptr++;
        *cptr++ = COLXS;
    }
}

outrow(rbuf,gbuf,bbuf,n)
unsigned short *rbuf, *gbuf, *bbuf;
int n;
{
    swizzleColr(rbuf,gbuf,bbuf,outbuf,n);
    gambs_colrs((COLR *)outbuf,n);
    fwrite(outbuf,4,n,ofile);		/* not run-length encoded */
}

/************************ IMPORTED CODE ***************************/

/* The following was imported from ray/src/common/header.c of Radiance 2.2 */

/*
 *  header.c - routines for reading and writing information headers.
 *
 *	8/19/88
 *
 *  printargs(ac,av,fp) print an argument list to fp, followed by '\n'
 *  isformat(s)		returns true if s is of the form "FORMAT=*"
 *  formatval(r,s)	copy the format value in s to r
 *  fputformat(s,fp)	write "FORMAT=%s" to fp
 *  getheader(fp,f,p)	read header from fp, calling f(s,p) on each line
 *  checkheader(i,p,o)	check header format from i against p and copy to o
 *
 *  To copy header from input to output, use getheader(fin, fputs, fout)
 */

char  FMTSTR[] = "FORMAT=";


printargs(ac, av, fp)		/* print arguments to a file */
int  ac;
char  **av;
register FILE  *fp;
{
	int  quote;

	while (ac-- > 0) {
		if (index(*av, ' ') != NULL) {		/* quote it */
			if (index(*av, '\'') != NULL)
				quote = '"';
			else
				quote = '\'';
			putc(quote, fp);
			fputs(*av++, fp);
			putc(quote, fp);
		} else
			fputs(*av++, fp);
		putc(' ', fp);
	}
	putc('\n', fp);
}


fputformat(s, fp)		/* put out a format value */
char  *s;
FILE  *fp;
{
	fputs(FMTSTR, fp);
	fputs(s, fp);
	putc('\n', fp);
}

/* End of included header.c code */

/* The following is imported from ray/src/common/colrops.c in Radiance 2.2 */

static BYTE	g_mant[256], g_nexp[256];

#ifndef pow
extern double	pow();
#endif


setcolrinv(f, a2)		/* set inverse brightness correction */
double	(*f)();
double	a2;
{
	double	mult;
	register int	i, j;
					/* compute gamb -> colr mapping */
	i = 0;
	mult = 256.0;
	for (j = 255; j > 0; j--) {
		while ((g_mant[j] = mult * (*f)(j/256.0, a2)) < 128) {
			i++;
			mult *= 2.0;
		}
		g_nexp[j] = i;
	}
	g_mant[0] = 0;
	g_nexp[0] = COLXS;
	return(0);
}


setcolrgam(g)			/* set gamma conversion */
double	g;
{
	return(setcolrinv(pow, g));
}


gambs_colrs(scan, len)		/* convert gamma bytes to colr scanline */
register COLR	*scan;
int	len;
{
	register int	nexpo;

	if (g_mant == NULL | g_nexp == NULL)
		return(-1);
	while (len-- > 0) {
		nexpo = g_nexp[scan[0][RED]];
		if (g_nexp[scan[0][GRN]] < nexpo)
			nexpo = g_nexp[scan[0][GRN]];
		if (g_nexp[scan[0][BLU]] < nexpo)
			nexpo = g_nexp[scan[0][BLU]];
		if (nexpo < g_nexp[scan[0][RED]])
			scan[0][RED] = g_mant[scan[0][RED]]
					>> (g_nexp[scan[0][RED]]-nexpo);
		else
			scan[0][RED] = g_mant[scan[0][RED]];
		if (nexpo < g_nexp[scan[0][GRN]])
			scan[0][GRN] = g_mant[scan[0][GRN]]
					>> (g_nexp[scan[0][GRN]]-nexpo);
		else
			scan[0][GRN] = g_mant[scan[0][GRN]];
		if (nexpo < g_nexp[scan[0][BLU]])
			scan[0][BLU] = g_mant[scan[0][BLU]]
					>> (g_nexp[scan[0][BLU]]-nexpo);
		else
			scan[0][BLU] = g_mant[scan[0][BLU]];
		scan[0][EXP] = COLXS - nexpo;
		scan++;
	}
	return(0);
}

/* End of included file from colrops.c */
